from typing import Dict, Optional, Iterable

import networkx as nx
import rich

from rich.progress import TaskID, TextColumn, TimeElapsedColumn, BarColumn
from rich.rule import Rule
from rich.console import Group

from .recipe import Recipe
from .types import Status, EvaluateProgress


class FancyProgress(rich.progress.Progress):
    def __init__(self, graph: nx.DiGraph, statuses: Dict[Recipe, Status], target_recipe: Recipe,
                 console: Optional[rich.console.Console] = None) -> None:
        """
        Prepare a progress object for the provided execution graph and statuses. Once created, call "start()" to begin
        showing the output, and "stop" to finish the output.

        :param graph: The graph to visualize progress for
        :param statuses: A dictionary of statuses per recipe
        :param target_recipe: The recipe that is the final target of the progress (last to run)
        :param console: An optional console object to use for output
        """
        self._console = rich.console.Console() if console is None else console
        self._recipe_name = target_recipe.name

        # Define the columns to use for progress table
        super().__init__(TextColumn("[deep_sky_blue2]{task.description}"),
                         BarColumn(),
                         TextColumn("{task.completed}/{task.total} ({task.percentage}%)"),
                         TextColumn("•"),
                         TimeElapsedColumn(),
                         console=self._console, redirect_stdout=True, redirect_stderr=True)

        # Build the progress table by adding all required tasks sorted topographically (target recipe at the bottom)
        self._recipe_tasks: Dict[Recipe, TaskID] = {
            recipe: self.add_task(recipe.name, start=False, total=1, completed=0)
            for recipe in nx.topological_sort(graph)
        }

        # For all recipes that are already cached (Ok), mark them as completed from the beginning
        for recipe, task_id in self._recipe_tasks.items():
            if statuses[recipe] == Status.Ok:
                self.update(task_id, description=recipe.name + " [dim cyan](cached)[/dim cyan]", completed=1)
                self.stop_task(task_id)

    def get_renderables(self) -> Iterable[rich.console.RenderableType]:
        """
        Helper function used to render a horizontal rule with the target recipe name above the actual progress bars
        """
        rule = Rule(title=f"Brewing {self._recipe_name}")
        yield Group(rule, self.make_tasks_table(self.tasks))

    def __call__(self, evaluate_progress: EvaluateProgress, recipe: Recipe, units_total: int, units_done: int) -> None:
        """
        Callable provided to execution engine. Will be called whenever a recipe has started progress, made progress (in
        the case of a ForeachRecipe) or finished evaluation. Note that this will only be called from a single thread, so
        we don't have to worry about race conditions.

        :param evaluate_progress: The type of progress being reported
        :param recipe: The recipe for which the evaluation progress is being reported
        :param units_total: The total units of work (always 1 for regular recipes)
        :param units_done: The currently finished units of work
        """
        if evaluate_progress == EvaluateProgress.Started:
            self.start_task(self._recipe_tasks[recipe])
        elif evaluate_progress == EvaluateProgress.InProgress:
            self.update(self._recipe_tasks[recipe], total=units_total, completed=units_done)
        elif evaluate_progress == EvaluateProgress.Done:
            self.update(self._recipe_tasks[recipe], total=units_total, completed=units_done)
            self.stop_task(self._recipe_tasks[recipe])



import enum
from typing import Callable, TYPE_CHECKING


@enum.unique
class Status(enum.Enum):
    """
    Status of a Recipe denoting whether (re)evaluation is needed
    """
    Ok = 0  # Recipe does not need (re)evaluation

    # The following all require (re)evaluation
    IngredientDirty = 1  # One or more ingredients of the recipe have changed
    NotEvaluatedYet = 2  # Recipe has not been evaluated yet
    InputsChanged = 3  # One or more inputs to the recipe have changed
    OutputsInvalid = 4  # One or more outputs of the recipe have been changed externally
    BoundFunctionChanged = 5  # The function referenced by the recipe has changed
    CustomDirty = 6  # The recipe has been marked dirty through a custom cleanliness function


@enum.unique
class ProgressType(enum.Enum):
    """
    Supported ways of showing progress
    """
    NoProgress = "none"  # Just run functions and log execution to alkymi's log
    Fancy = "fancy"  # Show progress indicators during recipe evaluation

    def __str__(self):
        return self.value


@enum.unique
class EvaluateProgress(enum.Enum):
    Started = 0
    InProgress = 1
    Done = 2


if TYPE_CHECKING:
    from .recipe import Recipe

# The status of the evaluation, the recipe for which the progress is to be updated, the total and current units of work
ProgressCallback = Callable[[EvaluateProgress, "Recipe", int, int], None]



import asyncio
import subprocess
import sys
import threading
from typing import List, TextIO, Optional, TypeVar, Callable

from .logging import log


def call(args: List[str], echo_error_to_stream: Optional[TextIO] = sys.stderr,
         echo_output_to_stream: Optional[TextIO] = sys.stdout) -> subprocess.CompletedProcess:
    """
    Utility command to run a system command and return the result (stdout and stderr will also be printed to the debug
    log)

    :param args: The arguments representing the command to run, e.g. ["echo", "test"]
    :param echo_error_to_stream: A stream to which to echo the call's stderr on a non-zero exit code
    :param echo_output_to_stream: A stream to which to echo the call's stdout while the command is executing
    :return: The result of the execution as a subprocess.CompletedProcess instance
    """
    # If running through a Lab CLI, sys.stdout may have been redirected
    if echo_output_to_stream is not None and getattr(echo_output_to_stream, "name", None) == sys.stdout.name:
        echo_output_to_stream = sys.stdout

    # Buffer one line at a time if echoing live, otherwise just use the default
    buffer_size = 1 if echo_output_to_stream is not None else -1
    proc = subprocess.Popen(args, universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            bufsize=buffer_size)

    # Since we set these to PIPE, these should never be None
    assert proc.stdout is not None
    assert proc.stderr is not None

    stdout: str = ""
    if echo_output_to_stream is not None:
        # Print each line to the stream as it arrives
        while proc.poll() is None:
            line = proc.stdout.readline()
            echo_output_to_stream.write(line)
            stdout += line

        # Program has finished executing, check if any part of stdout still needs to be piped
        line = proc.stdout.readline()
        if line:
            echo_output_to_stream.write(line)
            stdout += line
    else:
        # Otherwise, simply wait for the command to finish and then grab stdout
        proc.wait()
        stdout = proc.stdout.read()

    stderr = proc.stderr.read()

    # Send to alkymi log
    log.debug(stdout)
    log.debug(stderr)

    # Always forward error messages to stderr if needed
    if echo_error_to_stream is not None and proc.returncode != 0:
        echo_error_to_stream.write(stderr)

    # Raise an error on failure
    if proc.returncode != 0:
        raise subprocess.CalledProcessError(proc.returncode, proc.args, stdout, stderr)
    return subprocess.CompletedProcess(proc.args, proc.returncode, stdout, stderr)


T = TypeVar('T')


def run_on_thread(func: Callable[..., T]) -> T:
    """
    Execute a function on a new thread and return the result - exceptions will be propagated by re-raising
    :param func: The function to run on the new thread
    :return: The result of the function call
    """

    # Mutable object used to store result
    result: List[T] = []
    ex = []

    def _call_and_set_result():
        nonlocal result
        try:
            result.append(func())
        except Exception as e:
            ex.append(e)

    t = threading.Thread(target=_call_and_set_result)
    t.start()
    t.join()

    # Check if an exception occurred, and reraise
    if len(ex) != 0:
        raise ex[0]

    # Ensure that result has been set, and return it
    assert len(result) == 1, "Function call should have stored return value in result variable"
    return result[0]


def check_current_thread_has_running_event_loop():
    """
    Check if the calling thread has an associated running event loop
    :return: True if the calling thread has a running event loop associated with it
    """
    try:
        asyncio.get_running_loop()
        return True
    except RuntimeError:
        return False
